import os

import boto3 as boto3
import pytest


@pytest.fixture
def lambda_environment():
    client = boto3.Session(profile_name='dev').client('sts')
    lambda_role = 'arn:aws:iam::726075243133:role/worker-leech-dev-2-Oyster-168YECYUUDQMK'
    response = client.assume_role(
        RoleArn=lambda_role,
        RoleSessionName='lambdaTest'
    )
    credentials = response['Credentials']
    os.environ['AWS_ACCESS_KEY_ID'] = credentials['AccessKeyId']
    os.environ['AWS_SECRET_ACCESS_KEY'] = credentials['SecretAccessKey']
    os.environ['AWS_SESSION_TOKEN'] = credentials['SessionToken']



class AlgObject(object):
    @classmethod
    def parse_json(cls, json_dict):
        raise NotImplementedError()

    @classmethod
    def from_json(cls, json_dict):
        try:
            return cls.parse_json(json_dict)
        except KeyError:
            return cls.parse_json({x.replace('_', '', 1): y for x, y in json_dict.items()})

    @property
    def blessing(self):
        return 'this object blessed by Algernon Moncrief'

    @property
    def to_json(self):
        return self.__dict__

    @property
    def obj_name(self):
        return self.__class__.__name__

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            test = self.__dict__ == other.__dict__
            return test
        return False

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))



import json

import boto3
from botocore.exceptions import ClientError


class Opossum:
    @classmethod
    def get_database_secrets(cls):
        return cls.get_secrets('db_creds')

    @classmethod
    def get_wet_secrets(cls):
        return cls.get_secrets('neptune_creds')

    @classmethod
    def get_gql_secret(cls):
        return cls.get_secrets('gql')

    @classmethod
    def get_untrustworthy_export_key(cls, domain_name):
        secret_name = '%s_Export_Key' % domain_name
        secret = cls.get_secrets(secret_name)
        return secret['key_value']

    @classmethod
    def get_untrustworthy_credentials(cls, id_source):
        secret_name = 'credible_fe_%s' % id_source
        secret = cls.get_secrets(secret_name)
        return secret

    @classmethod
    def get_trident_user_key(cls):
        secret_name = 'Trident_User_Key'
        secret = cls.get_secrets(secret_name)
        return secret['trident_user_key_id'], secret['trident_user_key']

    @classmethod
    def get_secrets(cls, secret_name):
        client = boto3.client('secretsmanager')
        try:
            get_secret_value_response = client.get_secret_value(
                SecretId=secret_name
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                print("The requested secret " + secret_name + " was not found")
            elif e.response['Error']['Code'] == 'InvalidRequestException':
                print("The request was invalid due to:", e)
            elif e.response['Error']['Code'] == 'InvalidParameterException':
                print("The request had invalid params:", e)
        else:
            # Decrypted secret using the associated KMS CMK
            # Depending on whether the secret was a string or binary, one of these fields will be populated
            if 'SecretString' in get_secret_value_response:
                return json.loads(get_secret_value_response['SecretString'])
            else:
                return get_secret_value_response['SecretBinary']


class SneakyKipper:
    task_key_aliases = {
        'pagination': 'alias/algPaginationKey'
    }

    def __init__(self, task_name):
        self._client = boto3.client('kms')
        self._key_alis = self.task_key_aliases[task_name]

    def encrypt(self, unencrypted_text, encryption_context):
        import base64
        response = self._client.encrypt(
            KeyId=self._key_alis,
            Plaintext=json.dumps(unencrypted_text),
            EncryptionContext=encryption_context
        )
        bit = base64.b64encode(response['CiphertextBlob'])
        return bit.decode()

    def decrypt(self, encrypted_text, encryption_context):
        import base64

        response = self._client.decrypt(
            CiphertextBlob=base64.b64decode(encrypted_text),
            EncryptionContext=encryption_context
        )
        return json.loads(response['Plaintext'].decode())



import datetime
import hashlib
import hmac
import json
import logging
import os
import urllib.parse
from typing import Dict

import requests


class GqlSigner:
    _region = os.getenv('AWS_REGION', 'us-east-1')
    _service = 'appsync'

    def __init__(self, gql_endpoint: str):
        self._host = gql_endpoint
        self._uri = '/graphql'
        self._method = 'POST'
        self._signed_headers = 'host;x-amz-date'
        self._algorithm = 'AWS4-HMAC-SHA256'
        self._access_key = os.getenv('AWS_ACCESS_KEY_ID', None)
        self._secret_key = os.getenv('AWS_SECRET_ACCESS_KEY', None)
        self._session_token = os.getenv('AWS_SESSION_TOKEN', None)
        self._credentials = f"Credentials={self._access_key}"
        self._request_url = f'https://{gql_endpoint}{self._uri}'

    def generate_headers(self, query, variables):
        t = datetime.datetime.utcnow()
        amz_date = t.strftime('%Y%m%dT%H%M%SZ')
        date_stamp = t.strftime('%Y%m%d')
        canonical_request = self._generate_canonical_request(amz_date, query, variables)
        credential_scope = self._generate_scope(date_stamp)
        string_to_sign = self._generate_string_to_sign(canonical_request, amz_date, credential_scope)
        signature = self._generate_signature(string_to_sign, date_stamp)
        headers = self._generate_headers(credential_scope, signature, amz_date)
        return headers

    def _generate_canonical_request(self, amz_date, query, variables):
        payload = {'query': query, 'variables': variables}
        canonical_headers = f'host:{self._host}\nx-amz-date:{amz_date}\n'
        payload_hash = hashlib.sha256(json.dumps(payload).encode('utf-8')).hexdigest()
        canonical_request = f"{self._method}\n{self._uri}\n\n{canonical_headers}\n{self._signed_headers}\n{payload_hash}"
        return canonical_request

    def _generate_string_to_sign(self, canonical_request, amz_date, scope):
        hash_request = hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
        return f"{self._algorithm}\n{amz_date}\n{scope}\n{hash_request}"

    def _generate_scope(self, date_stamp):
        return f"{date_stamp}/{self._region}/{self._service}/aws4_request"

    def _get_signature_key(self, date_stamp):
        k_date = self._sign(f'AWS4{self._secret_key}'.encode('utf-8'), date_stamp)
        k_region = self._sign(k_date, self._region)
        k_service = self._sign(k_region, self._service)
        k_signing = self._sign(k_service, 'aws4_request')
        return k_signing

    def _generate_signature(self, string_to_sign, date_stamp):
        signing_key = self._get_signature_key(date_stamp)
        signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()
        return signature

    def _generate_headers(self, credential_scope, signature, amz_date):
        credentials_entry = f'Credential={self._access_key}/{credential_scope}'
        headers_entry = f'SignedHeaders={self._signed_headers}'
        signature_entry = f'Signature={signature}'
        authorization_header = f"{self._algorithm} {credentials_entry}, {headers_entry}, {signature_entry}"
        headers = {
            'x-amz-date': amz_date,
            'Authorization': authorization_header,
            'Content-Type': "application/graphql"}
        if self._session_token:
            headers['X-Amz-Security-Token'] = self._session_token
        return headers

    @classmethod
    def _generate_request_parameters(cls, command):
        payload = {'gremlin': command}
        request_parameters = urllib.parse.urlencode(payload, quote_via=urllib.parse.quote)
        payload_hash = hashlib.sha256(''.encode('utf-8')).hexdigest()
        return payload_hash, request_parameters

    @classmethod
    def _sign(cls, key, message):
        return hmac.new(key, message.encode('utf-8'), hashlib.sha256).digest()


class GqlNotary:
    _region = os.getenv('AWS_REGION', 'us-east-1')
    _service = 'appsync'

    def __init__(self, gql_endpoint: str, session: requests.session() = None):
        if not session:
            session = requests.session()
        self._session = session
        self._gql_endpoint = gql_endpoint
        self._uri = '/graphql'
        self._request_url = f'https://{gql_endpoint}{self._uri}'

    def send(self, command: str, variables: Dict = None):
        if not variables:
            variables = {}
        signer = GqlSigner(self._gql_endpoint)
        headers = signer.generate_headers(command, variables)
        payload = {'query': command, 'variables': variables}
        if os.getenv('DEBUG', 'False') == 'True':
            headers = {'x-api-key': os.environ['GQL_API_KEY']}
        response = requests.post(self._request_url, headers=headers, json=payload)
        if response.status_code != 200:
            raise RuntimeError(f'error communicating with GQL API: {response.text}, '
                               f'command: {command}, variables: {variables}')
        return response.text



import logging
import os

from aws_xray_sdk.core import patch_all


def lambda_logged(lambda_function):
    def wrapper(*args):
        patch_all()
        event = args[0]
        context = args[1]
        root = logging.getLogger()
        if root.handlers:
            for handler in root.handlers:
                root.removeHandler(handler)
        log_level = logging.INFO
        if os.getenv('DEBUG', 'false').lower() == 'true':
            log_level = logging.DEBUG
        logging.basicConfig(format='[%(levelname)s] ||' +
                                   f'function_name:{context.function_name}|function_arn:{context.invoked_function_arn}|'
                                   f'request_id:{context.aws_request_id}' +
                                   '|| %(asctime)s %(message)s', level=log_level)
        logging.getLogger('aws_xray_sdk').setLevel(logging.WARNING)
        logging.getLogger('botocore').setLevel(logging.WARNING)
        results = lambda_function(event, context)
        return results
    return wrapper



from queue import Queue
from threading import Thread
from typing import Dict

import boto3


class Bullhorn:
    def __init__(self, topic_map: Dict[str, str], client=None, num_batch_threads: int = 25):
        if not client:
            client = boto3.client('sns')
        self._client = client
        self._topic_map = topic_map
        self._num_threads = num_batch_threads
        self._batch_mode = False

    def __enter__(self):
        self._workers = []
        self._batch = Queue()
        self._batch_mode = True
        for _ in range(self._num_threads):
            worker = Thread(target=self._batch_publish)
            worker.start()
            self._workers.append(worker)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type and not exc_val:
            self._shutdown_workers()
            self._workers = []
            self._batch_mode = False
            return True
        raise (exc_type(exc_val))

    @classmethod
    def retrieve(cls, resource=None, profile=None, num_batch_threads: int = 25):
        topic_map = {}
        if not resource:
            resource = boto3.resource('sns')
            if profile:
                resource = boto3.Session(profile_name=profile).resource('sns')
        topic_iterator = resource.topics.all()
        for entry in topic_iterator:
            attributes = entry.attributes
            topic_arn = attributes['TopicArn']
            display_name = attributes['DisplayName']
            if display_name:
                topic_map[display_name] = topic_arn
        client = boto3.client('sns')
        if profile:
            client = boto3.Session(profile_name=profile).client('sns')
        return cls(topic_map, client, num_batch_threads=num_batch_threads)

    @property
    def topic_map(self):
        return self._topic_map

    def _shutdown_workers(self):
        for _ in self._workers:
            self._batch.put(None)
        for worker in self._workers:
            worker.join()

    def find_task_arn(self,  task_name):
        return self._topic_map.get(task_name)

    def publish(self, message_subject: str, topic_arn: str, message_body: str):
        if self._batch_mode:
            self._batch.put({
                "Subject": message_subject,
                "TopicArn": topic_arn,
                "Message": message_body
            })
            return
        response = self._client.publish(
            TopicArn=topic_arn,
            Subject=message_subject,
            Message=message_body
        )
        return response['MessageId']

    def _batch_publish(self):
        client = boto3.client('sns')
        while True:
            order = self._batch.get()
            if order is None:
                return
            client.publish(**order)
            self._batch.task_done()



from unittest.mock import patch

import pytest

from algernon.aws import Bullhorn


@pytest.mark.bullhorn
class TestBullhorn:
    def test_bullhorn_retrieve(self):
        bullhorn = Bullhorn.retrieve(profile='dev')
        assert bullhorn

    def test_bullhorn_live_send(self):
        bullhorn = Bullhorn.retrieve(profile='dev')
        arn = 'arn:aws:sns:us-east-1:726075243133:test'
        bullhorn.publish('test', arn, f'hello world, {1}')

    def test_bullhorn_batch_live_send(self):
        bullhorn = Bullhorn.retrieve(profile='dev')
        arn = 'arn:aws:sns:us-east-1:726075243133:test'
        with bullhorn as batch:
            for _ in range(100):
                batch.publish('test', arn, f'hello world, {_}')

    def test_bullhorn_batch_send(self):
        with patch('botocore.client.BaseClient._make_api_call') as mock:
            bullhorn = Bullhorn({'test': 'test_arn'})
            with bullhorn as batch:
                for _ in range(100):
                    batch.publish('test', 'some_arn', f'hello world, {_}')
            assert mock.called
            assert mock.call_count == 100



import uuid

import rapidjson

from algernon.aws import Bullhorn, StoredData
from algernon.serializers import AlgJson


def _generate_callback(current_task_name: str, callback: str):
    if not callback:
        return None, False
    chain_links = callback.split('#')
    if len(chain_links) == 1:
        return callback, False
    if current_task_name not in chain_links:
        return chain_links[0], True
    for pointer, link in enumerate(chain_links):
        if link == current_task_name:
            try:
                return chain_links[pointer+1], True
            except IndexError:
                return None, False


def _callback(original_payload, bullhorn=None, result=None, ):
    callback_fn, chained = _generate_callback(original_payload['task_name'], original_payload.get('callback'))
    if not callback_fn:
        return bullhorn
    if bullhorn is None:
        bullhorn = Bullhorn.retrieve()
    topic_arn = bullhorn.find_task_arn(callback_fn)
    if result is None:
        result = {}
    if result:
        result = StoredData.from_object(uuid.uuid4(), result, full_unpack=True)
    msg = {'task_name': callback_fn, 'task_kwargs': result}
    if chained:
        msg['callback'] = original_payload.get('callback')
    bullhorn.publish('callback', topic_arn, AlgJson.dumps(msg))
    return bullhorn


def rebuild_event(original_event):
    return AlgJson.loads(AlgJson.dumps(original_event))


def queued(production_fn):
    def wrapper(*args):
        results = []
        event = args[0]
        bullhorn = None
        local_context = {'aws': args[1]}
        for entry in event['Records']:
            entry_body = rapidjson.loads(entry['body'])
            original_payload = rapidjson.loads(entry_body['Message'])
            try:
                result = production_fn(original_payload, local_context)
                results.append(result)
                bullhorn = _callback(original_payload, bullhorn, result)
            except PipelineTerminationException as e:
                result = AlgJson.dumps(e.terminal_results)
                results.append(result)
        return results

    return wrapper


class PipelineTerminationException(Exception):
    def __init__(self, terminal_results):
        self._terminal_results = terminal_results

    @property
    def terminal_results(self):
        return self._terminal_results



import importlib
import json
from datetime import datetime
from decimal import Decimal

import pytz
from jsonref import JsonRef
import rapidjson

from algernon.alg_obj import AlgObject


class ExplosionJson:
    @classmethod
    def loads(cls, json_string):
        return rapidjson.loads(json_string, object_hook=ExplosionDecoder.object_hook)


class AlgJson:
    @classmethod
    def loads(cls, json_string):
        return rapidjson.loads(json_string, object_hook=AlgDecoder.object_hook)

    @classmethod
    def dumps(cls, obj):
        return rapidjson.dumps(obj, default=AlgEncoder.default)


class AlgEncoder(json.JSONEncoder):
    @classmethod
    def default(cls, obj):
        if isinstance(obj, AlgObject):
            return {'_alg_class': type(obj).__name__, '_alg_module': str(obj.__module__), 'value': obj.to_json}
        if isinstance(obj, frozenset):
            list_value = [x for x in obj]
            return {'_alg_class': 'frozenset', 'value': list_value}
        if isinstance(obj, set):
            list_value = [x for x in obj]
            return {'_alg_class': 'set', 'value': list_value}
        if isinstance(obj, tuple):
            return {'_alg_class': 'tuple', 'value': obj}
        if isinstance(obj, datetime):
            if obj.tzinfo is None:
                return {'_alg_class': 'datetime', 'value': {'tz': None, 'timestamp': obj.timestamp()}}
            return {'_alg_class': 'datetime', 'value': {'tz': obj.tzinfo.zone, 'timestamp': obj.timestamp()}}
        if isinstance(obj, Decimal):
            return {'_alg_class': 'decimal', 'value': str(obj)}
        if isinstance(obj, JsonRef):
            subject = obj.__subject__
            return subject
        return super(AlgEncoder, cls()).default(obj)


class AlgDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(obj):
        if '_alg_class' not in obj:
            return obj
        alg_class = obj['_alg_class']
        obj_value = obj['value']
        if alg_class == 'frozenset':
            return frozenset(x for x in obj_value)
        if alg_class == 'set':
            return set(x for x in obj_value)
        if alg_class == 'tuple':
            return tuple(x for x in obj_value)
        if alg_class == 'datetime':
            tz_info = obj_value['tz']
            timestamp = obj_value['timestamp']
            if tz_info is None:
                return datetime.fromtimestamp(timestamp)
            return datetime.fromtimestamp(timestamp, pytz.timezone(tz_info))
        if alg_class == 'decimal':
            return Decimal(obj_value)
        if '_alg_module' not in obj:
            return obj
        alg_module = obj['_alg_module']
        host_module = importlib.import_module(alg_module)
        obj_class = getattr(host_module, alg_class, None)
        if obj_class is None:
            raise RuntimeError(f'alg_class: {alg_class}, in alg_module: {alg_module} could not be located')
        return obj_class.from_json(obj_value)


class GqlDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(obj):
        if '_alg_class' not in obj:
            return obj
        alg_class = obj['_alg_class']
        obj_value = obj['value']
        if alg_class == 'frozenset':
            return frozenset(x for x in obj_value)
        if alg_class == 'tuple':
            return tuple(x for x in obj_value)
        if alg_class == 'datetime':
            return str(obj_value)
        if alg_class == 'decimal':
            return Decimal(obj_value)
        if '_alg_module' not in obj:
            return obj
        alg_module = obj['_alg_module']
        host_module = importlib.import_module(alg_module)
        obj_class = getattr(host_module, alg_class, None)
        if obj_class is None:
            return obj
        alg_obj = obj_class.from_json(obj_value)
        return alg_obj.to_gql


class ExplosionDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(obj):
        if isinstance(obj, dict):
            if 'S' in obj:
                return str(obj['S'])
            if 'M' in obj:
                return obj['M']
            if 'BOOL' in obj:
                return bool(obj['BOOL'])
            if 'NULL' in obj:
                return None
            if 'N' in obj:
                return Decimal(obj['N'])
            if 'L' in obj:
                return list(obj['L'])
            if 'SS' in obj:
                return set(obj['SS'])
            return obj
        return obj


class TaskDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(obj):
        if '_alg_class' not in obj:
            return obj
        alg_class = obj['_alg_class']
        obj_value = obj['value']
        if 'S' in obj:
            return str(obj['S'])
        if 'M' in obj:
            return obj['M']
        if 'BOOL' in obj:
            return bool(obj['BOOL'])
        if 'NULL' in obj:
            return None
        if 'N' in obj:
            return Decimal(obj['N'])
        if 'L' in obj:
            return list(obj['L'])
        return obj



import pytest

from algernon.aws import SneakyKipper


@pytest.mark.squirrel
class TestSquirrel:
    def test_squirrel_decrypt(self):
        unencrypted = 'testing, testing, 123'
        encryption_context = {'username': 'AROA2SDLU7J62MMDYWKJB:jcubeta'}
        kipper = SneakyKipper('pagination')
        encrypted = kipper.encrypt(unencrypted, encryption_context)
        decrypted = SneakyKipper('pagination').decrypt(encrypted, encryption_context)
        assert decrypted == unencrypted



from datetime import datetime

from algernon.alg_obj import AlgObject


class TridentProperty(AlgObject):
    def __init__(self, property_label, property_value):
        self._property_label = property_label
        self._property_value = property_value

    def __len__(self):
        return 1

    def __str__(self):
        return str(self._property_value)

    def __int__(self):
        return int(self._property_value)

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            json_dict['property_label'],
            json_dict['property_value']
        )

    @property
    def to_gql(self):
        property_value = {
            '__typename': 'ObjectPropertyValue',
            'data_type': 'S',
            'property_value': self._property_value
        }
        if isinstance(self._property_value, int):
            property_value['data_type'] = 'I'
        if isinstance(self._property_value, datetime):
            property_value['data_type'] = 'DT'

        return {
            '__typename': 'ObjectProperty',
            'property_name': self._property_label,
            'property_value': property_value
        }

    @property
    def label(self):
        return self._property_label

    @property
    def value(self):
        return self._property_value


class TridentVertex(AlgObject):
    def __init__(self, vertex_id, vertex_label, vertex_properties=None):
        if not vertex_properties:
            vertex_properties = {}
        self._vertex_id = vertex_id
        self._vertex_label = vertex_label
        self._vertex_properties = vertex_properties

    def __len__(self):
        return 1

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            json_dict['vertex_id'], json_dict['vertex_label'], json_dict['vertex_properties']
        )

    @property
    def to_gql(self):
        gql = {
            'internal_id': self.vertex_id,
            'vertex_type': self.vertex_label,
            '__typename': 'Vertex'
        }
        vertex_properties = []
        for property_name, vertex_property in self._vertex_properties.items():
            vertex_properties.append(vertex_property[0])
        if vertex_properties:
            gql['vertex_properties'] = vertex_properties
        return gql

    @property
    def vertex_id(self):
        return self._vertex_id

    @property
    def vertex_label(self):
        return self._vertex_label

    @property
    def vertex_properties(self):
        return self._vertex_properties

    @property
    def object_properties(self):
        vertex_properties = self.vertex_properties
        vertex_properties.update({
            'internal_id': self.vertex_id,
            'object_type': self.vertex_label
        })
        return vertex_properties

    def __iter__(self):
        return iter(self._vertex_properties)

    def __getitem__(self, item):
        return self._vertex_properties[item]

    def keys(self):
        return self._vertex_properties.keys()

    def values(self):
        return self._vertex_properties.values()

    def items(self):
        return self._vertex_properties.items()


class TridentEdge(AlgObject):
    def __init__(self, internal_id, label, from_vertex, to_vertex):
        self._internal_id = internal_id
        self._label = label
        self._from_vertex = from_vertex
        self._to_vertex = to_vertex

    def __len__(self):
        return 1

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            json_dict['internal_id'], json_dict['label'],
            json_dict['from_vertex'], json_dict['to_vertex'],
        )

    @property
    def to_gql(self):
        return {
            'internal_id': self.internal_id,
            'edge_label': self.label,
            'from_vertex': self._from_vertex,
            'to_vertex': self._to_vertex,
            '__typename': 'Edge'
        }

    @property
    def internal_id(self):
        return self._internal_id

    @property
    def label(self):
        return self._label

    @property
    def in_id(self):
        return self._from_vertex.vertex_id

    @property
    def out_id(self):
        return self._to_vertex.vertex_id


class TridentPath(AlgObject):
    def __init__(self, labels, path_objects):
        self._labels = labels
        self._path_objects = path_objects

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            json_dict['labels'],
            json_dict['path_objects']
        )

    @property
    def labels(self):
        return self._labels

    @property
    def path_objects(self):
        return self._path_objects


class TridentEdgeConnection(AlgObject):
    def __init__(self, edges, token, more, page_info=None):
        if not page_info:
            page_info = TridentPageInfo(token, more)
        self._edges = edges
        self._token = token
        self._more = more
        self._page_info = page_info

    @classmethod
    def parse_json(cls, json_dict):
        return cls(json_dict['edges'], json_dict['token'], json_dict['more'], json_dict.get('page_info'))

    @property
    def page_info(self):
        return self._page_info

    @property
    def in_count(self):
        return len(self._edges['inbound'])

    @property
    def out_count(self):
        return len(self._edges['outbound'])

    @property
    def total_count(self):
        return len(self._edges['all'])

    @property
    def to_gql(self):
        return {
            '__typename': 'EdgeConnection',
            'edges': self._edges,
            'page_info': self._page_info,
            'total_count': self.total_count,
            'in_count': self.in_count,
            'out_count': self.out_count
        }


class TridentPageInfo(AlgObject):
    def __init__(self, token, more):
        self._token = token
        self._more = more

    @classmethod
    def parse_json(cls, json_dict):
        return cls(json_dict['token'], json_dict['more'])

    @property
    def to_gql(self):
        return {
            'token': self._token,
            'more': self._more,
            '__typename': 'PageInfo'
        }



import pytest

from algernon.aws.gql import GqlSigner, GqlNotary


@pytest.mark.gql
@pytest.mark.usefixtures('lambda_environment')
class TestGql:
    def test_gql_signer(self):
        signer = GqlSigner('mh5syterirdvzji7tdbrrmpe7m.appsync-api.us-east-1.amazonaws.com')
        header = signer.generate_headers('some_query', {'some_variable': 'some_value'})
        assert header

    def test_notary(self):
        endpoint = 'mh5syterirdvzji7tdbrrmpe7m.appsync-api.us-east-1.amazonaws.com'
        notary = GqlNotary(endpoint)
        query = """
            query getEvent($flow_id: String!, $token: String){
                listStateEntries(flow_id: $flow_id, nextToken: $token){
                    items{
                        state_id
                    }
                }
            }
        """
        variables = {'flow_id': 'some_flow_id'}
        results = notary.send(query, variables)
        assert results



import datetime
import hashlib
import hmac
import json
import os
import urllib.parse

import rapidjson
import requests

from algernon.aws.squirrel import Opossum
from algernon.aws.trident.trident_obj import TridentVertex, TridentEdge, TridentProperty, TridentPath


class TridentDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(obj):
        if '@type' not in obj:
            return obj
        obj_type = obj['@type']
        obj_value = obj['@value']
        if obj_type == 'g:T':
            if obj_value == 'id':
                return 'internal_id'
            if obj_value == 'label':
                return 'label'
        if obj_type == 'g:Int32':
            return int(obj_value)
        if obj_type == 'g:Int64':
            return int(obj_value)
        if obj_type == 'g:List':
            return obj_value
        if obj_type == 'g:Set':
            return set(obj_value)
        if obj_type == 'g:Date':
            return datetime.datetime.fromtimestamp(obj_value/1000)
        if obj_type == 'g:Map':
            created_map = {}
            i = 0
            while i < len(obj_value):
                created_map[obj_value[i]] = obj_value[i + 1]
                i += 2
            return created_map
        if obj_type == 'g:Vertex':
            try:
                trident_vertex = TridentVertex(obj_value['id'], obj_value['label'], obj_value['properties'])
                return trident_vertex
            except KeyError:
                return TridentVertex(obj_value['id'], obj_value['label'])
        if obj_type == 'g:Edge':
            from_vertex = TridentVertex(obj_value['inV'], obj_value['inVLabel'])
            to_vertex = TridentVertex(obj_value['outV'], obj_value['outVLabel'])
            return TridentEdge(obj_value['id'], obj_value['label'],
                               from_vertex, to_vertex)
        if obj_type == 'g:VertexProperty':
            return TridentProperty(obj_value['label'], obj_value['value'])
        if obj_type == 'g:Path':
            return TridentPath(obj_value['labels'], obj_value['objects'])
        return obj


class TridentNotary:
    _region = os.getenv('AWS_REGION', 'us-east-1')
    _service = 'neptune-db'

    def __init__(self, neptune_endpoint, session=None):
        if neptune_endpoint is None:
            raise RuntimeError('must specify neptune endpoint when calling the trident_notary')
        if not session:
            session = requests.session()
        self._session = session
        self._neptune_endpoint = neptune_endpoint
        self._uri = '/gremlin/'
        self._method = 'POST'
        self._host = neptune_endpoint + ':8182'
        self._signed_headers = 'host;x-amz-date'
        self._algorithm = 'AWS4-HMAC-SHA256'
        access_key = os.getenv('AWS_ACCESS_KEY_ID', None)
        secret_key = os.getenv('AWS_SECRET_ACCESS_KEY', None)
        self._session_token = os.getenv('AWS_SESSION_TOKEN', None)
        if access_key is None or secret_key is None:
            access_key, secret_key = Opossum.get_trident_user_key()
        self._access_key = access_key
        self._secret_key = secret_key
        self._credentials = f"Credentials={self._access_key}"
        self._request_url = f'https://{neptune_endpoint}:8182{self._uri}'

    @classmethod
    def get_for_writer(cls, **kwargs):
        endpoint = kwargs.get('graph_db_endpoint', os.getenv('GRAPH_DB_ENDPOINT', None))
        return cls(endpoint)

    @classmethod
    def get_for_reader(cls, **kwargs):
        endpoint = kwargs.get('graph_db_reader_endpoint', os.getenv('GRAPH_DB_READER_ENDPOINT', None))
        return cls(endpoint)

    def send(self, command):
        t = datetime.datetime.utcnow()
        amz_date = t.strftime('%Y%m%dT%H%M%SZ')
        date_stamp = t.strftime('%Y%m%d')
        canonical_request, request_parameters = self._generate_canonical_request(amz_date, command)
        credential_scope = self._generate_scope(date_stamp)
        string_to_sign = self._generate_string_to_sign(canonical_request, amz_date, credential_scope)
        signature = self._generate_signature(string_to_sign, date_stamp)
        headers = self._generate_headers(credential_scope, signature, amz_date)
        get_results = self._session.post(self._request_url, headers=headers, json=request_parameters)
        if get_results.status_code != 200:
            raise RuntimeError(f'error passing command to remote database: {get_results.text}, command: {command}')
        response_json = rapidjson.loads(get_results.text)
        results = response_json['result']['data']
        results = rapidjson.loads(rapidjson.dumps(results), object_hook=TridentDecoder.object_hook)
        return results

    def _generate_canonical_request(self, amz_date, command):
        payload = rapidjson.dumps({'gremlin': command})
        canonical_headers = f'host:{self._host}\nx-amz-date:{amz_date}\n'
        payload_hash = hashlib.sha256(payload.encode('utf-8')).hexdigest()
        canonical_request = f"{self._method}\n{self._uri}\n\n{canonical_headers}\n{self._signed_headers}\n{payload_hash}"
        return canonical_request, payload

    def _generate_string_to_sign(self, canonical_request, amz_date, scope):
        hash_request = hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
        return f"{self._algorithm}\n{amz_date}\n{scope}\n{hash_request}"

    def _generate_scope(self, date_stamp):
        return f"{date_stamp}/{self._region}/{self._service}/aws4_request"

    def _get_signature_key(self, date_stamp):
        k_date = self._sign(f'AWS4{self._secret_key}'.encode('utf-8'), date_stamp)
        k_region = self._sign(k_date, self._region)
        k_service = self._sign(k_region, self._service)
        k_signing = self._sign(k_service, 'aws4_request')
        return k_signing

    def _generate_signature(self, string_to_sign, date_stamp):
        signing_key = self._get_signature_key(date_stamp)
        signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()
        return signature

    def _generate_headers(self, credential_scope, signature, amz_date):
        credentials_entry = f'Credential={self._access_key}/{credential_scope}'
        headers_entry = f'SignedHeaders={self._signed_headers}'
        signature_entry = f'Signature={signature}'
        authorization_header = f"{self._algorithm} {credentials_entry}, {headers_entry}, {signature_entry}"
        headers = {'x-amz-date': amz_date, 'Authorization': authorization_header}
        if self._session_token:
            headers['X-Amz-Security-Token'] = self._session_token
        return headers

    @classmethod
    def _generate_request_parameters(cls, command):
        payload = {'gremlin': command}
        request_parameters = urllib.parse.urlencode(payload, quote_via=urllib.parse.quote)
        payload_hash = hashlib.sha256(''.encode('utf-8')).hexdigest()
        return payload_hash, request_parameters

    @classmethod
    def _sign(cls, key, message):
        return hmac.new(key, message.encode('utf-8'), hashlib.sha256).digest()



import json
import logging
import os
from copy import deepcopy
from datetime import datetime

import boto3
from botocore.exceptions import ClientError
from algernon.alg_obj import AlgObject
from algernon.serializers import AlgJson, AlgEncoder


class StoredData(AlgObject):
    def __init__(self, data_name, data_string, bucket_name=None, folder_name=None, timestamp=None, full_unpack=False,
                 is_stored=None):
        if not bucket_name:
            bucket_name = os.environ['ALGERNON_BUCKET_NAME']
        if not folder_name:
            folder_name = os.getenv('CACHE_FOLDER_NAME', 'cache')
        if not timestamp:
            timestamp = str(datetime.utcnow().timestamp())
        self._data_name = data_name
        self._data_string = data_string
        self._bucket_name = bucket_name
        self._folder_name = folder_name
        self._timestamp = timestamp
        self._full_unpack = full_unpack
        self._is_stored = is_stored

    @property
    def to_json(self):
        if not self.check:
            self.store()
        return {'pointer': self.pointer}

    @property
    def check(self):
        if self._is_stored is True:
            return True
        resource = boto3.resource('s3')
        object_resource = resource.Object(self._bucket_name, self.data_key)
        try:
            object_resource.load()
        except ClientError as e:
            return int(e.response['Error']['Code']) != 404
        return True

    @property
    def pointer(self):
        return f'{self._bucket_name}#{self.data_key}'

    @property
    def data_key(self):
        return f'{self._folder_name}/{self._data_name}!{self._timestamp}.json'

    @property
    def data_string(self):
        return self._data_string

    @property
    def full_unpack(self):
        return self._full_unpack

    @classmethod
    def retrieve(cls, pointer):
        pointer_parts = pointer.split('#')
        key_parts = pointer_parts[1].split('!')
        folder_data_name = key_parts[0].split('/')
        folder_name = folder_data_name[0]
        data_name = folder_data_name[1]
        timestamp = key_parts[1].replace('.json', '')
        resource = boto3.resource('s3')
        stored_object = resource.Object(pointer_parts[0], pointer_parts[1]).get()
        string_body = stored_object['Body'].read()
        body = AlgJson.loads(string_body)
        cls_args = {
            'data_name': data_name,
            'data_string': body['data_string'],
            'bucket_name': pointer_parts[0],
            'folder_name': folder_name,
            'timestamp': timestamp,
            'full_unpack': body['full_unpack'],
            'is_stored': True
        }
        return cls(**cls_args)

    @classmethod
    def parse_json(cls, json_dict):
        pointer = json_dict['pointer']
        stored_data = cls.retrieve(pointer)
        if stored_data.full_unpack:
            return stored_data.data_string
        return stored_data

    @classmethod
    def from_object(cls, data_name, alg_object, full_unpack=False):
        if isinstance(alg_object, StoredData):
            logging.debug(f'tried to store a StoredData object within another stored data object, that was naughty, '
                          f'you will go to jail now. jk, '
                          f'we just bypassed the upload and handed the original back: {alg_object}')
            return alg_object
        return cls(data_name, alg_object, full_unpack=full_unpack)

    def store(self):
        if self.check:
            raise RuntimeError('can not overwrite stored data')
        self._overwrite_store()

    def _overwrite_store(self):
        resource = boto3.resource('s3')
        body = {'data_string': self._data_string, 'full_unpack': self._full_unpack}
        try:
            body_string = AlgJson.dumps(body)
        except TypeError:
            body_string = json.dumps(body, cls=AlgEncoder)
        resource.Object(self._bucket_name, self.data_key).put(Body=body_string)
        self._is_stored = True
        return self.pointer

    def __str__(self):
        return self.pointer

    def merge(self, other_stored_data):
        current_data = deepcopy(self._data_string)
        other_data_string = other_stored_data.data_string
        if other_data_string is None:
            return False
        for data_name, data_entry in other_stored_data.data_string.items():
            if data_name in current_data:
                current_data_entry = current_data[data_name]
                if not isinstance(current_data_entry, list):
                    current_data[data_name] = [current_data_entry]
                if data_entry not in current_data[data_name]:
                    current_data[data_name].append(data_entry)
                continue
            current_data[data_name] = data_entry
        if current_data == self._data_string:
            return False
        self._data_string = current_data
        self._overwrite_store()
        return True



from algernon.aws.trident.connections import TridentNotary


class TridentDriver:
    def __init__(self, **kwargs):
        self._read_notary = kwargs.get('read_notary', TridentNotary.get_for_reader(**kwargs))
        self._write_notary = kwargs.get('write_notary', TridentNotary.get_for_writer(**kwargs))
        self._batch_mode = False

    def get(self, internal_id):
        command = "g.V('%s')" % internal_id
        return self.execute(command, True)

    def execute(self, query_text, read_only=False):
        if self._batch_mode is True:
            self._batch_commands.append(query_text)
            return
        notary = self._write_notary
        if read_only:
            notary = self._read_notary
        results = notary.send(query_text)
        return results

    def __enter__(self):
        self._batch_commands = []
        self._batch_mode = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type and not exc_val:
            self._batch_mode = False
            if self._batch_commands:
                commands = ';'.join(self._batch_commands)
                self.execute(commands)
            self._batch_commands = []
            return True
        raise (exc_type(exc_val))



import io
import os
import sys

from setuptools import setup, Command, find_packages
from shutil import rmtree


NAME = 'moncrief'
DESCRIPTION = 'everything to run a project in a distributed and serverless fashion'
URL = 'https://github.com/AlgernonSolutions/algernon'
EMAIL = 'jcubeta@algernon.solutions'
AUTHOR = 'algernon_solutions/jcubeta'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '2.3.13'

REQUIRED = [
    'aws-xray-sdk',
    'boto3',
    'botocore',
    'certifi',
    'chardet',
    'docutils',
    'future',
    'idna',
    'jmespath',
    'jsonpickle',
    'jsonref',
    'python-dateutil',
    'python-rapidjson',
    'pytz',
    'requests',
    's3transfer',
    'six',
    'urllib3',
    'wrapt'
]

EXTRAS = {}

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        # self.status('Building Source and Wheel (universal) distribution…')
        print('Building Source and Wheel (universal) distribution…')
        os.system('python setup.py sdist bdist_wheel --universal')

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        #self.status('Pushing git tags…')
        #os.system('git tag v{0}'.format(about['__version__']))
        #os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=about['__version__'],
    url=URL,
    license='GNU Affero General Public License v3.0',
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRED,
    cmdclass={
        'upload': UploadCommand,
    }
)
